<?php
namespace Civi\Api4\Action\ContactCategory;

use Civi\Api4\Contact;
use Civi\Api4\Activity;
use Civi\Api4\ContactCategory;
use Civi\Api4\ContactCategoryDefinition;
use Civi\Api4\Group;
use CRM_Contactcats_ExtensionUtil as E;
use Civi\Test\CiviEnvBuilder;
use Civi\Test\HeadlessInterface;
use Civi\Test\HookInterface;
use Civi\Test\TransactionalInterface;

/**
 * Check our stats are predicatable.
 *
 * Tips:
 *  - With HookInterface, you may implement CiviCRM hooks directly in the test class.
 *    Simply create corresponding functions (e.g. "hook_civicrm_post(...)" or similar).
 *  - With TransactionalInterface, any data changes made by setUp() or test****() functions will
 *    rollback automatically -- as long as you don't manipulate schema or truncate tables.
 *    If this test needs to manipulate schema or truncate tables, then either:
 *       a. Do all that using setupHeadless() and Civi\Test.
 *       b. Disable TransactionalInterface, and handle all setup/teardown yourself.
 *
 * @method void assertNotEmpty()
 * @method void assertEquals()
 * @method void assertMatchesRegularExpression()
 * @group headless
 */
class GetFlowsTest extends \PHPUnit\Framework\TestCase implements HeadlessInterface, HookInterface, TransactionalInterface {

  /**
   * Setup used when HeadlessInterface is implemented.
   *
   * Civi\Test has many helpers, like install(), uninstall(), sql(), and sqlFile().
   *
   * @link https://github.com/civicrm/org.civicrm.testapalooza/blob/master/civi-test.md
   *
   * @return \Civi\Test\CiviEnvBuilder
   *
   * @throws \CRM_Extension_Exception_ParseException
   */
  public function setUpHeadless(): CiviEnvBuilder {
    return \Civi\Test::headless()
      ->installMe(__DIR__)
      ->apply();
  }

  public function setUp():void {
    parent::setUp();
  }

  public function tearDown():void {
    parent::tearDown();
  }

  /**
   * Example: Test that a version is returned.
   */
  // public function testWellFormedVersion():void {
  //   $this->assertNotEmpty(E::SHORT_NAME);
  //   $this->assertMatchesRegularExpression('/^([0-9\.]|alpha|beta)*$/', \CRM_Utils_System::version());

  /**
   * }
   */
  public function testGetFlows():void {
    // Create groups for our cats.
    [$amazingGroupID, $mehGroupID] = Group::save(FALSE)
      ->setRecords([
        ['name' => 'amazing', 'frontend_title' => 'amazing'],
        ['name' => 'meh', 'frontend_title' => 'meh'],
      ])->execute()->column('id');

    // Create cats.
    [$amazingCatID, $mehCatID] = ContactCategoryDefinition::save(FALSE)
      ->setRecords([
        [
          'label' => 'Amazing',
          'search_type:name' => 'group',
          'search_data' => ['group_id' => $amazingGroupID],
          'presentation_order' => 1,
          'execution_order' => 1,
          'color' => '#ff0000',
        ],
        [
          'label' => 'Meh',
          'search_type:name' => 'group',
          'search_data' => ['group_id' => $mehGroupID],
          'presentation_order' => 2,
          'execution_order' => 2,
          'color' => '#fff000',
        ],
      ])
      ->execute()->column('id');

    [$ctID, $ctID2] = Contact::save(FALSE)
      ->setDefaults(['contact_type' => 'Individual'])
      ->setRecords([
        ['display_name' => 'Wilma'],
        ['display_name' => 'Fred'],
      ])->execute()->column('id');

    // Put them in meh category. This is not required for the test,
    // but makes the data realistic.
    $ccIDs = ContactCategory::save(FALSE)
      ->setRecords([
        ['id' => $ctID, 'category_definition_id' => $mehCatID],
        ['id' => $ctID2, 'category_definition_id' => $mehCatID],
      ])->execute()->column('id');
    $this->assertEquals([$ctID, $ctID2], $ccIDs);
    // print "ccs\n" . implode("\n", array_map('json_encode', CRM_Core_DAO::executeQuery("SELECT * from civicrm_contact_category cc")->fetchAll())) . "\n---\n";

    // Create two activities for ct1 and just one for ct2
    $activityIDs = Activity::save(FALSE)
      ->setDefaults([
        'activity_type_id:name' => 'changed_contact_category',
        'target_contact_id' => $ctID,
        'source_contact_id' => $ctID,
      ])
      ->setRecords([
        [
          'activity_date_time' => '2025-01-01',
          'Category_changes.new_category_id' => $amazingCatID,
          'subject' => "Fake subject: $ctID change to amazing from nothing",
        ],
        [
          'activity_date_time' => '2025-01-01',
          'Category_changes.new_category_id' => $mehCatID,
          'target_contact_id' => $ctID2,
          'subject' => "Fake subject: $ctID2 change to meh from nothing",
        ],
        [
          'activity_date_time' => '2025-02-01',
          'Category_changes.new_category_id' => $mehCatID,
          'subject' => "Fake subject: $ctID change to meh from amazing",
        ],
      ])->execute()->column('id');

    // ================
    // At 3 Jan, we had 1 amazing, 1 meh
    // Later, 1 amazing moved to meh.
    // Flows should return non-changes, too.
    $flows = ContactCategory::getFlows()
      ->setStartDate('2025-01-03')
      ->execute()->getArrayCopy();
    $this->assertEquals([
      ['from_category_id' => $amazingCatID, 'to_category_id' => $mehCatID, 'contact_count' => 1],
      ['from_category_id' => $mehCatID, 'to_category_id' => $mehCatID, 'contact_count' => 1],
    ], $flows);

  }

}
